#include "app.h"
#include <math.h>
#include <iomanip>
#include <windows.h>
#include "glfw.h"


// Convert from degrees to radians
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}


Application::Application( const string &contentDir )
{
	for( unsigned int i = 0; i < 320; ++i ) _keys[i] = false;

	//_x = 0; _y = 1; _z = 10; _rx = 7; _ry = 15; 
	

	_x=-4.642200f; 
	_y=0.512304f; 
	_z=0.203532f; 
	_rx=-6.200000f; 
	_ry=-86.699989f;
	
	_velocity = 5.0f;
	_curFPS = 30;

	_freeze = false; _showStats = true; _debugViewMode = false; _wireframeMode = false;
	_animTime = 0; _weight = 1.0f;
	_cam = 0; _light = 0;

	_contentDir = contentDir;
	_flashGUIRes = 0;
	_flashCubeRes = 0;
	_deleteFlash = false;
	_flashCounter = 0;

}


bool Application::reload_resources() {

	H3DRes reshandle=0;
	int restype;

	while(reshandle=h3dGetNextResource(H3DResTypes::Undefined,reshandle)) {
		restype=h3dGetResType(reshandle);
		h3dUnloadResource(reshandle);
	}

	h3dutLoadResourcesFromDisk("./content");

	H3DRes matRes = h3dFindResource( H3DResTypes::Material, "pipelines/postHDR.material.xml" );

	// hdrParams: exposure, brightpass threshold, brightpass offset (see shader for description)
	h3dSetMaterialUniform( matRes, "hdrParams", 10.0f, 0.5f, 0.08f, 0 );
	return true;
}

void Application::dumpMessages() {

	int level;
	float timestamp;
	static char *errors[]={ "ERROR", "WARN", "INFO", "DEBUG" };

	while(1) {
		const char *mes = h3dGetMessage(&level,&timestamp);
		if(*mes==0) break;
		printf("%s: %s\n",errors[level-1],mes);
	}
}

char *Application::flashCallback(char *command,char *args, void *userdata) {
	printf("command: %s args: %s userdata:%x\n",command,args,userdata);
	Application *app=(Application *)userdata;
	if(strcmp(command,"menu") == 0) {
		if(strcmp(args,"continue")==0) { app->_deleteFlash = true; } 
		if(strcmp(args,"exit") == 0) exit(0);
	}
	if(strcmp(command,"brightness") == 0) {
		float val = atoi(args)/100.0f;
		h3dSetNodeParamF( app->_light, H3DLight::ColorF3, 0, val );
		h3dSetNodeParamF( app->_light, H3DLight::ColorF3, 1, val );
		h3dSetNodeParamF( app->_light, H3DLight::ColorF3, 2, val );
	}

	return "okidoki";
}

bool Application::init()
{	
	// Initialize engine
	if( !h3dInit() )
	{	
		dumpMessages();
		return false;
	}

	// Set options
	h3dSetOption( H3DOptions::LoadTextures, 1 );
	h3dSetOption( H3DOptions::TexCompression, 0 );
	h3dSetOption( H3DOptions::FastAnimation, 0 );
	h3dSetOption( H3DOptions::MaxAnisotropy, 8 );
	h3dSetOption( H3DOptions::ShadowMapSize, 2048 );

	// Add resources
	// Pipelines
	_hdrPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/hdr.pipeline.xml", 0 );
	_forwardPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/forward.pipeline.xml", 0 );
	// Font
	_fontMatRes = h3dAddResource( H3DResTypes::Material, "overlays/font.material.xml", 0 );
	_panelMatRes = h3dAddResource( H3DResTypes::Material, "overlays/panel.material.xml", 0 );

	// material that will contain the flash texture, in this case an overlay
	_flashGUIMatRes = h3dAddResource( H3DResTypes::Material, "overlays/flash.material.xml", 0 );
	
	H3DRes cuberes = h3dAddResource( H3DResTypes::SceneGraph, "cube/cube.scene.xml", 0 );

	//	trick to forward all windowevents from main window
	//	h3dfForwardWindowEvents(_flashRes,h3dfGetAppWindow());

	// Load resources
	h3dutLoadResourcesFromDisk( _contentDir.c_str() );

	// Add camera
	_cam = h3dAddCameraNode( H3DRootNode, "Camera", _hdrPipeRes );
	h3dSetNodeTransform( _cam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );

	H3DNode cube = h3dAddNodes( H3DRootNode, cuberes );
	h3dSetNodeTransform( cube, 0, 0, 0, 0, 0, 0, 1.0f, 1.0f, 1.0f );

	// Add light source
	_light = h3dAddLightNode( H3DRootNode, "Light1", 0, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform( _light, 1, 20, 1, -90, 0, 0, 1, 1, 1 );
	h3dSetNodeParamF( _light, H3DLight::RadiusF,0, 20 );
	h3dSetNodeParamF( _light, H3DLight::FovF,0, 90 );
	h3dSetNodeParamI( _light, H3DLight::ShadowMapCountI, 1 );
	h3dSetNodeParamF( _light, H3DLight::ShadowMapBiasF,0, 0.01f );

	// Customize post processing effects
	H3DRes matRes = h3dFindResource( H3DResTypes::Material, "pipelines/postHDR.material.xml" );
	// hdrParams: exposure, brightpass threshold, brightpass offset (see shader for description)
	h3dSetMaterialUniform( matRes, "hdrParams", 10.0f, 0.5f, 0.08f, 0 );

	_flashGUIRes = h3dfCreateTexture("flash",1024,768);
	h3dSetResParamI(_flashGUIMatRes, H3DMatRes::SamplerElem, 0, H3DMatRes::SampTexResI, _flashGUIRes);


	H3DRes cubeMatRes = h3dFindResource( H3DResTypes::Material, "cube/cube.material.xml" );
	_flashCubeRes = h3dGetResParamI(cubeMatRes, H3DMatRes::SamplerElem, 0, H3DMatRes::SampTexResI);
  	h3dfCreate(_flashCubeRes,140,150,380,340,NULL);
	h3dfLoadFromFile(_flashCubeRes,"FlexExample2/bin/FlexExample2.swf");

	toggleMenu();
	return true;
}

void Application::toggleMenu() {
	if(h3dfGetHorde3DFlash(_flashGUIRes)) {
		this->_deleteFlash = true;
	}else{
		h3dfCreate(_flashGUIRes,0,0,1024,768,NULL);
		h3dfSetCallBack(_flashGUIRes,flashCallback,this);
		h3dfLoadFromFile(_flashGUIRes,"FlexExample/bin/FlexExample.swf");
	}
}

void Application::mainLoop( float fps )
{
	_curFPS = fps;

	
	h3dSetOption( H3DOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	h3dSetOption( H3DOptions::WireframeMode, _wireframeMode ? 1.0f : 0.0f );
		
	
	if( _showStats )
	{
		h3dutShowFrameStats( _fontMatRes, _panelMatRes, 1 );
	}

	h3dfRender();

	if(h3dfGetHorde3DFlash(_flashGUIRes)) {
		if(_deleteFlash) {
			h3dfRemoveTexture(_flashGUIRes);
			_deleteFlash = false;
		}else{
			// render Flash textures
	
			h3dShowOverlay( 0, 0, 0, 1,  // top-left
							0, 1, 0, 0,  // bottom-left
							1, 1, 1, 0,	 // bottom-right
							1, 0, 1, 1,  // top-right
							1, 1, 1, 1, 
							_flashGUIMatRes, 7 );
		}
	}else{

		keyHandler();
		// Set camera parameters
		h3dSetNodeTransform( _cam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );
		h3dSetNodeTransform( _light, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );

	}
	// Render scene
	h3dRender( _cam );


	// Finish rendering of frame
	h3dFinalizeFrame();

	// Remove all overlays
	h3dClearOverlays();

	// Write all mesages to log file
	dumpMessages();

}

void Application::release()
{
	// Release engine
	h3dfRemoveTexture(_flashGUIRes);
	h3dRelease();
}

void Application::resize( int width, int height )
{
	// Resize viewport
	h3dSetupViewport( 0, 0, width, height, true );
	
	// Set virtual camera parameters
	h3dSetupCameraView( _cam, 45.0f, (float)width / height, 0.01f, 10000.0f );
}


void Application::keyPressEvent( int key )
{
	if( key == 32 )		// Space
		_freeze = !_freeze;

	if( key == 260 )	// F3
	{
		if( h3dGetNodeParamI( _cam, H3DCamera::PipeResI ) == _hdrPipeRes )
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _forwardPipeRes );
		else
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _hdrPipeRes );
	}
	
	if( key == 262) // F5 
		reload_resources();

	if( key == 264 ){	// F7
		_debugViewMode = !_debugViewMode;
	}
	if( key == 265 )	// F8
		_wireframeMode = !_wireframeMode;
	
	if( key == 266 )	// F9
		_showStats = !_showStats;

}


void Application::keyHandler()
{
	float curVel = _velocity / _curFPS;
	
	if( _keys[287] ) curVel *= 5;	// LShift
	
	if( _keys['W'] )
	{
		// Move forward
		_x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y -= sinf( -degToRad( _rx ) ) * curVel;
		_z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys['S'] )
	{
		// Move backward
		_x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y += sinf( -degToRad( _rx ) ) * curVel;
		_z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys['A'] )
	{
		// Strafe left
		_x += sinf( degToRad( _ry - 90) ) * curVel;
		_z += cosf( degToRad( _ry - 90 ) ) * curVel;
	}

	if( _keys['D'] )
	{
		// Strafe right
		_x += sinf( degToRad( _ry + 90 ) ) * curVel;
		_z += cosf( degToRad( _ry + 90 ) ) * curVel;
	}

	if( _keys['1'] )
	{
		// Change blend weight
		_weight += 2 / _curFPS;
		if( _weight > 1 ) _weight = 1;
	}
	
	if( _keys['2'] )
	{
		// Change blend weight
		_weight -= 2 / _curFPS;
		if( _weight < 0 ) _weight = 0;
	}
}


void Application::mouseMoveEvent( float dX, float dY )
{
	if(h3dfGetHorde3DFlash(_flashGUIRes)) {
		int x,y;
		glfwGetMousePos(&x,&y);
		char val[256];
		_itoa_s(int(x),val,256,10);

		h3dfCallFunction(_flashGUIRes,"compass",val);
		// use modulo because x and y can get higher than screen size
		h3dfSendMouseMove(_flashGUIRes,UINT(x) % 1024,UINT(y) % 768);
	}else{
	
		// Look left/right
		_ry -= dX / 100 * 30;
		
		// Loop up/down but only in a limited range
		_rx += dY / 100 * 30;
		if( _rx > 90 ) _rx = 90; 
		if( _rx < -90 ) _rx = -90;


	}
}

void Application::mouseButtonEvent( int type, int up_or_down)
{
	if(h3dfGetHorde3DFlash(_flashGUIRes)) {
		if(up_or_down == GLFW_PRESS)
			h3dfSendMouseDown(_flashGUIRes);
		else if(up_or_down == GLFW_RELEASE)
			h3dfSendMouseUp(_flashGUIRes); 
	}
}

