// Horde3D Render flash to Texture by Niek Albers niek@daansystems.com
// v0.3

#include <atlbase.h>
#include <map>
#include "Horde3D.h"
#include <rpcsal.h>

using namespace std;

#import "PROGID:ShockwaveFlash.ShockwaveFlash" 
using namespace ShockwaveFlashObjects;

typedef char* (*H3DFLASH_CALLBACK_PROC)(char *cmd, char *args, void *userdata);

#define CHECKHR(x) if((hr=x)!=0) errorMsg(hr,__LINE__);
#define MAX_EXTERNALINTERFACE_STRLEN 256

class Horde3DFlash : 
	public IOleClientSite, 
	public IOleInPlaceSiteWindowless, 
	public IOleInPlaceFrame,
	public IStorage, 
	public _IShockwaveFlashEvents {

public:
	int refCount;
	RECT targetRect;
	RECT flashRect;
	HDC hdcCompatible;  
	HBITMAP hBitmap;
	UINT *frame_buffer;
	DWORD conPointID;
	H3DRes textureRes;
	HRESULT hr;
	LONG_PTR wndProc;
	H3DFLASH_CALLBACK_PROC callback;
	void *userdata;
	bstr_t callFunctionResult;
	int posx;
	int posy;
	bool mouseDown;
	BITMAPINFO binfo;
	int texWidth;
	int texHeight;
	bool useRead;

	bool dirty;
	CComQIPtr<IShockwaveFlash> flash;
	CComQIPtr<IOleObject> flashOleObject;
	CComQIPtr<IViewObjectEx> flashViewObject;
	CComQIPtr<IOleInPlaceObjectWindowless> flashInPlaceObjWindowless;
	CComQIPtr<IPersistStreamInit> persistStream;
	CComQIPtr<IConnectionPoint> conPoint;

	void errorMsg(HRESULT hr, int line) {
		char error[1024]; 
		_snprintf_s(error,1024,"ERROR: %x line:%d\n",hr,line); 
		::MessageBox(NULL,error,"Horde3DFlash ERROR",MB_OK); 
	}

	static H3DRes createTexture(char *name, int width, int height) {
		return h3dCreateTexture(name, width,height, H3DFormats::TEX_BGRA8, H3DResFlags::NoTexMipmaps | H3DResFlags::TexDynamic);
	}

	Horde3DFlash(H3DRes textureRes, int left, int top, int right, int bottom, char *flashVars = NULL) {
		this->refCount = 0;
		this->callback = NULL;
		this->dirty = false;
		this->posx = 0;
		this->posy = 0;
		this->mouseDown = false;
		this->hr=0;
		this->textureRes = textureRes;
		this->texWidth = h3dGetResParamI(textureRes, H3DTexRes::ImageElem, 0, H3DTexRes::ImgWidthI);
		this->texHeight = h3dGetResParamI(textureRes, H3DTexRes::ImageElem, 0, H3DTexRes::ImgHeightI);

		OleInitialize(NULL);
		CHECKHR(OleCreate(__uuidof(ShockwaveFlash), IID_IOleObject, OLERENDER_DRAW, 0,  this, this, (void **) &flashOleObject));
		CHECKHR(OleSetContainedObject(flashOleObject, TRUE));

		flash = flashOleObject;
		flashViewObject = flashOleObject;
		flashInPlaceObjWindowless = flashOleObject;

		CComQIPtr<IConnectionPointContainer> conPointContainer = flash;
		CHECKHR(conPointContainer->FindConnectionPoint(__uuidof(_IShockwaveFlashEvents), &conPoint));
		CHECKHR(conPoint->Advise((_IShockwaveFlashEvents *)this, &conPointID));
		//	CHECKHR(flash->put_AllowScriptAccess(L"always"));
		//	CHECKHR(flash->DisableLocalSecurity());
		CHECKHR(flash->put_WMode(L"transparent"));
		//	CHECKHR(flash->put_WMode(L"opaque"));
		CHECKHR(flash->put_FlashVars(bstr_t(flashVars)));

		targetRect.left = left;
		targetRect.top = top;
		targetRect.right = right;
		targetRect.bottom = bottom;

		flashRect.left = 0;
		flashRect.top = 0;
		flashRect.right = targetRect.right - targetRect.left;
		flashRect.bottom = targetRect.bottom - targetRect.top;

		useRead = (targetRect.left > 0 || targetRect.top > 0 || targetRect.right < this->texWidth || targetRect.bottom < this->texHeight);

		CHECKHR(flashInPlaceObjWindowless->SetObjectRects(&flashRect,&flashRect));
		CHECKHR(flashOleObject->DoVerb(OLEIVERB_INPLACEACTIVATE, NULL, this, 0, NULL, NULL));  

		CreateBuffer();
	}


	~Horde3DFlash() {
		this->conPoint->Unadvise(this->conPointID);
		this->flashOleObject->Close(OLECLOSE_NOSAVE); 
		::DeleteDC(hdcCompatible);
		::DeleteObject(hBitmap);
	}

	//IUnknown 
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void ** ppvObject)
	{
		if (IsEqualGUID(riid, IID_IUnknown))
			*ppvObject = (void*)(this);
		else if (IsEqualGUID(riid, IID_IOleInPlaceSite))
			*ppvObject = (void*)dynamic_cast<IOleInPlaceSite *>(this);
		else if (IsEqualGUID(riid, IID_IOleInPlaceSiteEx))
			*ppvObject = (void*)dynamic_cast<IOleInPlaceSiteEx *>(this);
		else if (IsEqualGUID(riid, IID_IOleInPlaceSiteWindowless))
			*ppvObject = (void*)dynamic_cast<IOleInPlaceSiteWindowless *>(this);
		else if (IsEqualGUID(riid, IID_IStorage))
			*ppvObject = (void*)dynamic_cast<IStorage *>(this);
		else if (IsEqualGUID(riid, __uuidof(_IShockwaveFlashEvents)))
			*ppvObject = (void*)dynamic_cast<_IShockwaveFlashEvents *>(this);
		else if (IsEqualGUID(riid, IID_IOleInPlaceFrame))
			*ppvObject = (void*)dynamic_cast<IOleInPlaceFrame *>(this);
		else
			*ppvObject = 0;
		if (!(*ppvObject))
			return E_NOINTERFACE; //if dynamic_cast returned 0
		refCount++;
		return S_OK;
	}

	ULONG STDMETHODCALLTYPE AddRef() { 
		return ++refCount; 
	}
	ULONG STDMETHODCALLTYPE Release() { return --refCount; }
	// IOleClientSite
	HRESULT STDMETHODCALLTYPE SaveObject( void) { return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE GetMoniker( 
		/* [in] */ DWORD dwAssign,
		/* [in] */ DWORD dwWhichMoniker,
		/* [out] */ __RPC__deref_out_opt IMoniker **ppmk) { *ppmk=NULL; return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE GetContainer( 
		/* [out] */ __RPC__deref_out_opt IOleContainer **ppContainer) { *ppContainer = 0; return E_NOINTERFACE; }
	HRESULT STDMETHODCALLTYPE ShowObject( void) {  return S_OK; }
	HRESULT STDMETHODCALLTYPE OnShowWindow( 
		/* [in] */ BOOL fShow) { return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE RequestNewObjectLayout( void) { return E_NOTIMPL; }

	// IOleInPlaceSite
	HRESULT STDMETHODCALLTYPE CanInPlaceActivate( void) { return S_OK; }
	HRESULT STDMETHODCALLTYPE OnInPlaceActivate( void) { return S_OK; }
	HRESULT STDMETHODCALLTYPE OnUIActivate( void) { return S_OK; }
	HRESULT STDMETHODCALLTYPE GetWindowContext( 
		/* [out] */ __RPC__deref_out_opt IOleInPlaceFrame **ppFrame,
		/* [out] */ __RPC__deref_out_opt IOleInPlaceUIWindow **ppDoc,
		/* [out] */ __RPC__out LPRECT lprcPosRect,
		/* [out] */ __RPC__out LPRECT lprcClipRect,
		/* [out][in] */ __RPC__inout LPOLEINPLACEFRAMEINFO lpFrameInfo) {

			*ppFrame = (IOleInPlaceFrame *)this;
			*ppDoc = NULL;

			lpFrameInfo->fMDIApp = FALSE;
			lpFrameInfo->hwndFrame = NULL;
			lpFrameInfo->haccel = NULL;
			lpFrameInfo->cAccelEntries = 0;


			*lprcPosRect = flashRect;
			*lprcClipRect = flashRect;
			return S_OK;
	}

	HRESULT STDMETHODCALLTYPE Scroll( 
		/* [in] */ SIZE scrollExtant)  { return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE OnUIDeactivate( 
		/* [in] */ BOOL fUndoable) { return S_OK; }
	HRESULT STDMETHODCALLTYPE OnInPlaceDeactivate( void) { return S_OK; }
	HRESULT STDMETHODCALLTYPE DiscardUndoState( void) { return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE DeactivateAndUndo( void) { return E_NOTIMPL; }
	HRESULT STDMETHODCALLTYPE OnPosRectChange( 
		/* [in] */ __RPC__in LPCRECT lprcPosRect) { return S_OK; }
	/* [input_sync] */ HRESULT STDMETHODCALLTYPE GetWindow( 
		/* [out] */ __RPC__deref_out_opt HWND *phwnd) { 
			*phwnd = 0; 
			return E_FAIL; 
			//		return S_OK;
	}
	HRESULT STDMETHODCALLTYPE ContextSensitiveHelp( 
		/* [in] */ BOOL fEnterMode)  { return E_NOTIMPL; }

	//IOleInPlaceSiteEx

	virtual HRESULT STDMETHODCALLTYPE OnInPlaceActivateEx( 
		/* [out] */ __RPC__out BOOL *pfNoRedraw,
		/* [in] */ DWORD dwFlags) { if (pfNoRedraw) *pfNoRedraw = FALSE; return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE OnInPlaceDeactivateEx( 
		/* [in] */ BOOL fNoRedraw) { return S_FALSE; }
	virtual HRESULT STDMETHODCALLTYPE RequestUIActivate( void) { return S_FALSE; }

	// IOleInPlaceSiteWindowless

	virtual HRESULT STDMETHODCALLTYPE CanWindowlessActivate( void) {  return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE GetCapture( void) { return S_FALSE; }
	virtual HRESULT STDMETHODCALLTYPE SetCapture( 
		/* [in] */ BOOL fCapture) { 
			return S_FALSE; 
	}
	virtual HRESULT STDMETHODCALLTYPE GetFocus( void) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE SetFocus(/* [in] */ BOOL fFocus) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE GetDC( 
		/* [unique][in] */ __RPC__in_opt LPCRECT pRect,
		/* [in] */ DWORD grfFlags,
		/* [out] */ __RPC__deref_out_opt HDC *phDC) { return S_FALSE; }
	virtual HRESULT STDMETHODCALLTYPE ReleaseDC( 
		/* [in] */ __RPC__in HDC hDC) { return S_FALSE; }
	virtual HRESULT STDMETHODCALLTYPE InvalidateRect( 
		/* [unique][in] */ __RPC__in_opt LPCRECT pRect,
		/* [in] */ BOOL fErase) {
			this->dirty = true;
			return S_OK;
	}
	virtual HRESULT STDMETHODCALLTYPE InvalidateRgn( 
		/* [in] */ __RPC__in HRGN hRGN,
		/* [in] */ BOOL fErase)  { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE ScrollRect( 
		/* [in] */ INT dx,
		/* [in] */ INT dy,
		/* [in] */ __RPC__in LPCRECT pRectScroll,
		/* [in] */ __RPC__in LPCRECT pRectClip) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE AdjustRect( 
		/* [out][in] */ __RPC__inout LPRECT prc) { return S_FALSE; }
	virtual HRESULT STDMETHODCALLTYPE OnDefWindowMessage( 
		/* [in] */ UINT msg,
		/* [in] */ WPARAM wParam,
		/* [in] */ LPARAM lParam,
		/* [out] */ __RPC__out LRESULT *plResult) { return S_FALSE; }

	//IDispatch Impl
	HRESULT STDMETHODCALLTYPE GetTypeInfoCount(UINT __RPC_FAR *pctinfo)
	{
		return E_NOTIMPL;
	}

	HRESULT STDMETHODCALLTYPE GetTypeInfo( 
		/* [in] */ UINT iTInfo,
		/* [in] */ LCID lcid,
		/* [out] */ ITypeInfo __RPC_FAR *__RPC_FAR *ppTInfo)
	{
		return E_NOTIMPL;
	}
	HRESULT STDMETHODCALLTYPE GetIDsOfNames( 
		/* [in] */ REFIID riid,
		/* [size_is][in] */ LPOLESTR __RPC_FAR *rgszNames,
		/* [in] */ UINT cNames,
		/* [in] */ LCID lcid,
		/* [size_is][out] */ DISPID __RPC_FAR *rgDispId)
	{
		return E_NOTIMPL;
	}

	HRESULT STDMETHODCALLTYPE Invoke( 
		/* [in] */ DISPID dispIdMember,
		/* [in] */ REFIID riid,
		/* [in] */ LCID lcid,
		/* [in] */ WORD wFlags,
		/* [out][in] */ DISPPARAMS __RPC_FAR *pDispParams,
		/* [out] */ VARIANT __RPC_FAR *pVarResult,
		/* [out] */ EXCEPINFO __RPC_FAR *pExcepInfo,
		/* [out] */ UINT __RPC_FAR *puArgErr)
	{

		//	ATLTRACE("invoke: %x\n",dispIdMember);
		switch (dispIdMember)   
		{   
		case 0x7a6:            
			break;   
		case 0x96:   // fscommand        
			if ((pDispParams->cArgs == 2) &&   
				(pDispParams->rgvarg[0].vt == VT_BSTR) &&   
				(pDispParams->rgvarg[1].vt == VT_BSTR))   
			{   
				//			printf("1:%S 2:%S\n",pDispParams->rgvarg[1].bstrVal, pDispParams->rgvarg[0].bstrVal);

				char command[MAX_EXTERNALINTERFACE_STRLEN];
				char args[MAX_EXTERNALINTERFACE_STRLEN];

				_snprintf_s(command,MAX_EXTERNALINTERFACE_STRLEN,"%S",pDispParams->rgvarg[1].bstrVal);
				_snprintf_s(args,MAX_EXTERNALINTERFACE_STRLEN,"%S",pDispParams->rgvarg[0].bstrVal);

				if(this->callback) {
					char *result = this->callback(command, args, this->userdata);
				}
			}   
			break;   

		case 0xc5:  // ExternalInterface.call(..)
			if(this->callback) {
				char xml[1024];
				_snprintf_s(xml,1024,"%S",pDispParams->rgvarg[0].bstrVal);
				char command[MAX_EXTERNALINTERFACE_STRLEN];
				char args[MAX_EXTERNALINTERFACE_STRLEN];

				sscanf_s(xml, "<invoke name=\"%[^\"]\" returntype=\"xml\"><arguments><string>%[^<]</string></arguments></invoke>", command,MAX_EXTERNALINTERFACE_STRLEN,args,MAX_EXTERNALINTERFACE_STRLEN);
				char *result=this->callback(command,args,this->userdata);
				CHECKHR(flash->SetReturnValue(bstr_t("<string>")+bstr_t(result)+bstr_t("</string>")));
			}
			break;
		case DISPID_READYSTATECHANGE:                      
			break;   
		default:               
			return DISP_E_MEMBERNOTFOUND;   
		}    

		return S_OK;
	}

	// IStorage
	virtual HRESULT STDMETHODCALLTYPE CreateStream( 
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsName,
		/* [in] */ DWORD grfMode,
		/* [in] */ DWORD reserved1,
		/* [in] */ DWORD reserved2,
		/* [out] */ __RPC__deref_out_opt IStream **ppstm) { return E_NOTIMPL; }
	virtual /* [local] */ HRESULT STDMETHODCALLTYPE OpenStream( 
		/* [string][in] */ const OLECHAR *pwcsName,
		/* [unique][in] */ void *reserved1,
		/* [in] */ DWORD grfMode,
		/* [in] */ DWORD reserved2,
		/* [out] */ IStream **ppstm) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE CreateStorage( 
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsName,
		/* [in] */ DWORD grfMode,
		/* [in] */ DWORD reserved1,
		/* [in] */ DWORD reserved2,
		/* [out] */ __RPC__deref_out_opt IStorage **ppstg) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE OpenStorage( 
		/* [string][unique][in] */ __RPC__in_opt_string const OLECHAR *pwcsName,
		/* [unique][in] */ __RPC__in_opt IStorage *pstgPriority,
		/* [in] */ DWORD grfMode,
		/* [unique][in] */ __RPC__deref_opt_in_opt SNB snbExclude,
		/* [in] */ DWORD reserved,
		/* [out] */ __RPC__deref_out_opt IStorage **ppstg) { return E_NOTIMPL; }
	virtual /* [local] */ HRESULT STDMETHODCALLTYPE CopyTo( 
		/* [in] */ DWORD ciidExclude,
		/* [size_is][unique][in] */ const IID *rgiidExclude,
		/* [annotation][unique][in] */ 
		__RPC__in_opt  SNB snbExclude,
		/* [unique][in] */ IStorage *pstgDest) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE MoveElementTo( 
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsName,
		/* [unique][in] */ __RPC__in_opt IStorage *pstgDest,
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsNewName,
		/* [in] */ DWORD grfFlags)  { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE Commit( 
		/* [in] */ DWORD grfCommitFlags) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE Revert( void) { return E_NOTIMPL; }
	virtual /* [local] */ HRESULT STDMETHODCALLTYPE EnumElements( 
		/* [in] */ DWORD reserved1,
		/* [size_is][unique][in] */ void *reserved2,
		/* [in] */ DWORD reserved3,
		/* [out] */ IEnumSTATSTG **ppenum) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE DestroyElement( 
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsName) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE RenameElement( 
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsOldName,
		/* [string][in] */ __RPC__in_string const OLECHAR *pwcsNewName) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE SetElementTimes( 
		/* [string][unique][in] */ __RPC__in_opt_string const OLECHAR *pwcsName,
		/* [unique][in] */ __RPC__in_opt const FILETIME *pctime,
		/* [unique][in] */ __RPC__in_opt const FILETIME *patime,
		/* [unique][in] */ __RPC__in_opt const FILETIME *pmtime) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE SetClass( 
		/* [in] */ __RPC__in REFCLSID clsid) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE SetStateBits( 
		/* [in] */ DWORD grfStateBits,
		/* [in] */ DWORD grfMask) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE Stat( 
		/* [out] */ __RPC__out STATSTG *pstatstg,
		/* [in] */ DWORD grfStatFlag) { return E_NOTIMPL; }

	//IOleInPlaceFrame

	virtual /* [input_sync] */ HRESULT STDMETHODCALLTYPE GetBorder( 
		/* [out] */ __RPC__out LPRECT lprectBorder) { return E_NOTIMPL; }
	virtual /* [input_sync] */ HRESULT STDMETHODCALLTYPE RequestBorderSpace( 
		/* [unique][in] */ __RPC__in_opt LPCBORDERWIDTHS pborderwidths) { return E_NOTIMPL; }
	virtual /* [input_sync] */ HRESULT STDMETHODCALLTYPE SetBorderSpace( 
		/* [unique][in] */ __RPC__in_opt LPCBORDERWIDTHS pborderwidths) { return E_NOTIMPL; }
	virtual HRESULT STDMETHODCALLTYPE SetActiveObject( 
		/* [unique][in] */ __RPC__in_opt IOleInPlaceActiveObject *pActiveObject,
		/* [unique][string][in] */ __RPC__in_opt_string LPCOLESTR pszObjName) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE InsertMenus( 
		/* [in] */ __RPC__in HMENU hmenuShared,
		/* [out][in] */ __RPC__inout LPOLEMENUGROUPWIDTHS lpMenuWidths) { return E_NOTIMPL; }
	virtual /* [input_sync] */ HRESULT STDMETHODCALLTYPE SetMenu( 
		/* [in] */ __RPC__in HMENU hmenuShared,
		/* [in] */ __RPC__in HOLEMENU holemenu,
		/* [in] */ __RPC__in HWND hwndActiveObject) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE RemoveMenus( 
		/* [in] */ __RPC__in HMENU hmenuShared) { return E_NOTIMPL; }
	virtual /* [input_sync] */ HRESULT STDMETHODCALLTYPE SetStatusText( 
		/* [unique][in] */ __RPC__in_opt LPCOLESTR pszStatusText) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE EnableModeless( 
		/* [in] */ BOOL fEnable) { return S_OK; }
	virtual HRESULT STDMETHODCALLTYPE TranslateAccelerator( 
		/* [in] */ __RPC__in LPMSG lpmsg,
		/* [in] */ WORD wID){ return E_NOTIMPL; }

	HRESULT OnReadyStateChange (
		long newState ) { return S_OK; }
	HRESULT OnProgress (
		long percentDone ) { return S_OK; }
	HRESULT FSCommand (
		_bstr_t command,
		_bstr_t args ) { 
			return S_OK; 
	}
	HRESULT FlashCall (
		_bstr_t request ) { return S_OK; }

	void* CreateBuffer(){   
		hdcCompatible = CreateCompatibleDC(NULL);    

		memset(&binfo,0, sizeof(binfo));   
		binfo.bmiHeader.biSize = sizeof(binfo);   
		binfo.bmiHeader.biPlanes = 1;   
		binfo.bmiHeader.biBitCount = 32;   
		binfo.bmiHeader.biCompression = BI_RGB;   
		binfo.bmiHeader.biWidth = this->flashRect.right;   
		binfo.bmiHeader.biHeight = this->flashRect.bottom;   

		hBitmap = ::CreateDIBSection(hdcCompatible, &binfo, DIB_RGB_COLORS, (void **)&frame_buffer, NULL, 0);   
		::SelectObject(hdcCompatible, hBitmap);      
		::SetMapMode(hdcCompatible, MM_TEXT);   

		return frame_buffer;   
	}   

	void render() {
		if(this->dirty) {
			UINT bytes = flashRect.right * flashRect.bottom * sizeof(UINT);
			RECTL rectl;
			rectl.top = 0;
			rectl.left = 0;
			rectl.right = flashRect.right;
			rectl.bottom = flashRect.bottom;
			memset(frame_buffer,0,bytes);
			CHECKHR(flashViewObject->Draw(DVASPECT_CONTENT, -1, NULL, NULL, NULL, this->hdcCompatible, &rectl, NULL, NULL, NULL));			
			// writeonly is flash texture is the full quad
			UINT* data = (UINT*)h3dMapResStream(this->textureRes, H3DTexRes::ImageElem, 0, H3DTexRes::ImgPixelStream, useRead, true );
			// copy buffer with flipped y coordinates
		 	UINT i = 0;
			for(LONG y=targetRect.bottom-1;y>=targetRect.top;y--) {
				UINT offsetY = y*this->texWidth;
				for(LONG x=targetRect.left;x<targetRect.right;x++) {
					data[x+offsetY] = frame_buffer[i];
					i++;
				}
			}
		//	printf("copied: %d pixels total: %d\n",i,flashRect.right*flashRect.bottom); 
			h3dUnmapResStream(this->textureRes);
			this->dirty = false;
		}
	}

	static LRESULT WINAPI MainWndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam ) {
		LONG_PTR t=::GetWindowLongPtr(hWnd, GWLP_USERDATA);	
		LRESULT result;		
		((Horde3DFlash *)t)->flashInPlaceObjWindowless->OnWindowMessage(msg,wParam,lParam,&result);
		return ::CallWindowProc((WNDPROC)((Horde3DFlash *)t)->wndProc,hWnd,msg,wParam,lParam);
	}

	BOOL forwardWindowEvents(HWND wnd) {
		this->wndProc = ::GetWindowLongPtr(wnd, GWLP_WNDPROC);
		::SetWindowLongPtr(wnd, GWLP_WNDPROC, (LONG_PTR) MainWndProc);
		::SetWindowLongPtr(wnd, GWLP_USERDATA, (LONG_PTR) this);
		return TRUE;
	}

	LRESULT sendMessage(  /* [in] */ UINT msg,
		/* [in] */ WPARAM wParam,
		/* [in] */ LPARAM lParam) {
			LRESULT result;		
			CHECKHR(flashInPlaceObjWindowless->OnWindowMessage(msg,wParam,lParam,&result));
			return result;
	}

	BOOL sendMouseMove(int x,int y) {
		this->posx = x;
		this->posy = y;

		LRESULT result;		
		CHECKHR(flashInPlaceObjWindowless->OnWindowMessage(WM_MOUSEMOVE, this->mouseDown ? MK_LBUTTON : 0,MAKELPARAM(this->posx,this->posy),&result));
		return result == S_OK;
	}

	BOOL sendMouseDown() {
		if(this->mouseDown) return FALSE;
		this->mouseDown = true;
		LRESULT result;
		HRESULT hr=flashInPlaceObjWindowless->OnWindowMessage(WM_LBUTTONDOWN,MK_LBUTTON,MAKELPARAM(this->posx,this->posy),&result);
		return result == S_OK;
	}

	BOOL sendMouseUp() {
		if(!this->mouseDown) return FALSE;
		this->mouseDown = false;
		LRESULT result;
		CHECKHR(flashInPlaceObjWindowless->OnWindowMessage(WM_LBUTTONUP,0,MAKELPARAM(this->posx,this->posy),&result));
		return result == S_OK;
	}

	BOOL loadFromMemory(BYTE *data, UINT length) {
		HGLOBAL hMem = GlobalAlloc(GMEM_MOVEABLE,length+8);

		LPBYTE pImage = (LPBYTE)::GlobalLock(hMem);
		//header magic
		UINT *headerpos = (UINT *)pImage;
		*headerpos = 0x55665566;
		*(headerpos+1) = length;
		::CopyMemory(pImage+8,data,length);
		::GlobalUnlock(hMem);

		CComPtr<IStream> spStream;
		CHECKHR(CreateStreamOnHGlobal(hMem,FALSE,&spStream));

		persistStream = flashOleObject;
		persistStream->InitNew();
		CHECKHR(persistStream->Load(spStream));
		::GlobalFree(hMem);
		return S_OK;
	}

	BOOL loadFromFile(char *filename) {
		TCHAR szDirectory[MAX_PATH];
		::GetCurrentDirectory(MAX_PATH, szDirectory);
		CComBSTR path(szDirectory);
		path.Append("\\");
		path.Append(filename);
		CHECKHR(flash->put_Movie(path));
		return hr == S_OK;
	}

	BOOL setCallBack(H3DFLASH_CALLBACK_PROC proc, void *userdata = NULL) {
		this->callback = proc;
		this->userdata = userdata;
		return proc ? TRUE : FALSE;
	}

	static BOOL CALLBACK EnumWindowProc(HWND hwnd, LPARAM lParam)
	{
		HINSTANCE hinst=(HINSTANCE)GetModuleHandle(NULL);

		if((HINSTANCE)::GetWindowLongPtr(hwnd, GWL_HINSTANCE)==hinst &&
			::IsWindowVisible(hwnd))
		{
			*((HWND *)lParam)=hwnd;
			return FALSE;
		}
		else
			return TRUE;
	}

	char *callFunction(char *name, char *args = NULL) {
		bstr_t name2 = bstr_t("<invoke name=\"") + bstr_t(name) + bstr_t("\" returntype=\"xml\"><arguments><string>") + bstr_t(args) + bstr_t("</string></arguments></invoke>"); 

		this->callFunctionResult = this->flash->CallFunction(name2);

		char *result = this->callFunctionResult;

		// very simple xml parser ;)
		char *lasttag = strrchr(result,'<');
		*lasttag=0;
		return strchr(result,'>')+1;
	}

	BOOL setVariable(char *name,char *value) {

		CHECKHR(flash->SetVariable(bstr_t(name),bstr_t(value)));
		return hr == S_OK;

	}

	BOOL play() {
		return this->flash->Play() == S_OK;
	}

	BOOL stop() {
		return this->flash->Stop() == S_OK;
	}
};


#ifndef DLLEXP
#define DLLEXP extern "C" __declspec( dllexport )
#endif

map<H3DRes,Horde3DFlash*> textures;

DLLEXP Horde3DFlash *h3dfGetHorde3DFlash( H3DRes textureRes ) {
	map<H3DRes,Horde3DFlash*>::iterator i = textures.find(textureRes);
	if(i == textures.end()) return NULL;
	return i->second;
}

DLLEXP H3DRes h3dfCreateTexture( char *name, int width, int height)
{
	return Horde3DFlash::createTexture(name,width,height);
}

DLLEXP BOOL h3dfCreate(H3DRes textureRes, int left, int top, int right, int bottom, char *flashVars) {
	if(!textureRes) return FALSE;
	Horde3DFlash *flash = new Horde3DFlash(textureRes, left, top, right, bottom,flashVars);	
	textures[textureRes] = flash;
	return TRUE;
}

DLLEXP BOOL h3dfRemoveTexture( H3DRes textureRes )
{
	Horde3DFlash *flash = textures[textureRes];
	if(!flash) return FALSE;
	textures.erase(textureRes);
	delete flash;
	return TRUE;
}

DLLEXP BOOL h3dfLoadFromFile( H3DRes textureRes, char *filename )
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->loadFromFile(filename);
}

DLLEXP BOOL h3dfLoadFromMemory( H3DRes textureRes, void *data, unsigned int length)
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->loadFromMemory((BYTE *)data,length);
}

DLLEXP BOOL h3dfSendMouseMove( H3DRes textureRes, int x, int y)
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->sendMouseMove(x,y);
}

DLLEXP BOOL h3dfSendMouseDown(H3DRes textureRes) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->sendMouseDown();
}

DLLEXP BOOL h3dfSendMouseUp(H3DRes textureRes) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->sendMouseUp();
}

DLLEXP LRESULT h3dfSendMessage( H3DRes textureRes, UINT msg, WPARAM wParam, LPARAM lParam) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->sendMessage(msg,wParam,lParam);
}

DLLEXP BOOL h3dfRender() 
{
	map<H3DRes,Horde3DFlash*>::iterator i = textures.begin();
	for( ; i != textures.end(); ++i ) {
		if(i->second) i->second->render();
	}
	return true;
}

DLLEXP BOOL h3dfForwardWindowEvents( H3DRes textureRes, HWND wnd) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->forwardWindowEvents(wnd);
}

DLLEXP BOOL h3dfSetVariable( H3DRes textureRes, char *name, char *value) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->setVariable(name,value);
}

DLLEXP char *h3dfCallFunction(H3DRes textureRes, char *name, char *args) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->callFunction(name,args);

}

DLLEXP BOOL h3dfPlay(H3DRes textureRes) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->play();
}

DLLEXP BOOL h3dfStop(H3DRes textureRes) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->stop();
}

DLLEXP BOOL h3dfSetCallBack(H3DRes textureRes, H3DFLASH_CALLBACK_PROC proc, void *userdata) 
{
	Horde3DFlash *flash = textures[textureRes];
	_ASSERT(flash);
	return flash->setCallBack(proc,userdata);
}

DLLEXP HWND h3dfGetAppWindow() {
	HWND wnd = NULL;
	::EnumWindows(Horde3DFlash::EnumWindowProc, (LPARAM)&wnd);
	return wnd;
}
